// Mark Kamichoff
// kamicm@rpi.edu
// Final Project

#include "defines.h"

// Some globals
int win_width=INIT_W;
int win_height=INIT_H;
GLUquadricObj* gluObj;
short debug=0;

// Inits the external object structure
external_objects extobj[NUMOBJS];

// Texture data, right now we load only one
int texture[1];

// Lights (-1 is off, 1 is on)
short lite0 = 1;
short lite1 = -1;
short lite2 = -1;
short lite3 = -1;

// misc -1 is keyb, 1 is mouse
GLint nNumPoints = 4;

// View Stuff (bool_shipview {-1=follow,1=lookfromcenter}
short bool_shipview = -1;

GLfloat follow_zoom=1.2;
short bool_follow_zoom=0;
short bool_angy=0;
short bool_movez=0;
short bool_movey=0;
short drawstyle = 0;

GLfloat ratio;
short orbit_xang_bool=0;
short orbit_yang_bool=0;
short orbit_zang_bool=0;
short orbit_zoom_bool=0;
GLfloat orbit_xang = 75;
GLfloat orbit_yang = 25;
GLfloat orbit_zoom = 2;

GLfloat gradual_x = -2.0;
GLfloat gradual_y = 0;
GLfloat gradual_z = -1.0;
GLfloat gradual_center_x = 0;
GLfloat gradual_center_y = -1.0;
GLfloat gradual_center_z = 0;

// Timing variables, need to be global
int frame=0, frame_time, timebase=0;
double fps=29.97;

// Selection variables
// Maximum of 32 ships, and { xpos,ypos,zpos,xang,yang,zang,special,visible,size }
// A value of -1 for selected_ship means nothing is selected and the camera
// view should be orbiting and walking
GLfloat shipdb[32][9];
short selected_ship=-1;

// Does time calculations (fps, etc...)
void timestuff(void)
{
	frame++;
	frame_time = glutGet(GLUT_ELAPSED_TIME);
	
	if (frame_time - timebase > 1000)
	{
		if(debug)
			printf("FPS:%4.2f\n", fps);
		fps=frame*1000.0/(frame_time-timebase);
		timebase = frame_time;		
		frame = 0;
	}
}

// Read in the selected models that will be used
void readextobjs(void)
{
	read_raw_obj("models/ywing.raw", 0);
	read_raw_obj("models/tiebomber.raw", 1);
	read_raw_obj("models/sd.raw", 2);
	read_raw_obj("models/falcon.raw", 3);
}

// Update the values when you move the camera or object
void updatevalues(void)
{
	// If we're orbiting and stuff
	if(selected_ship == -1)
	{
		if(orbit_xang_bool == 1)
			orbit_xang += 60/fps;
		else if(orbit_xang_bool == -1)
			orbit_xang -= 60/fps;
	
		if(orbit_yang_bool == 1)
			orbit_yang += 30/fps;
		else if(orbit_yang_bool == -1)
			orbit_yang -= 60/fps;
	
		if(orbit_zoom_bool == 1)
//			orbit_zoom += 0.01;
			orbit_zoom += 0.6/fps;
		else if(orbit_zoom_bool == -1)
//			orbit_zoom -= 0.01;
			orbit_zoom -= 0.6/fps;
	}
	else
	{
		if(bool_angy == 1)
			shipdb[selected_ship][4] += 60/fps;
		else if(bool_angy == -1)
			shipdb[selected_ship][4] -= 60/fps;

		if(bool_movez == 1)
		{
			shipdb[selected_ship][0] +=
				cos(PI*((int)(shipdb[selected_ship][4])%360)/180) * 0.6/fps;
			shipdb[selected_ship][2] +=
				sin(PI*((int)(shipdb[selected_ship][4])%360)/180) * 0.6/fps;
		}
		else if(bool_movez == -1)
		{
			shipdb[selected_ship][0] -=
				cos(PI*((int)(shipdb[selected_ship][4])%360)/180) * 0.6/fps;
			shipdb[selected_ship][2] -=
				sin(PI*((int)(shipdb[selected_ship][4])%360)/180) * 0.6/fps;
		}

		if(bool_movey == 1)
			shipdb[selected_ship][1] += 0.6/fps;
		else if(bool_movey == -1)
			shipdb[selected_ship][1] -= 0.6/fps;

		if(bool_follow_zoom == 1)
			follow_zoom += 0.4/fps;
		else if(bool_follow_zoom == -1)
		{
			if(follow_zoom<0)
				follow_zoom = 0;
			else
				follow_zoom -= 0.6/fps;
		}
	}

	// Update lighting
	if(lite0 == 1) glEnable(GL_LIGHT0);
	else glDisable(GL_LIGHT0);
	if(lite1 == 1) glEnable(GL_LIGHT1);
	else glDisable(GL_LIGHT1);
	if(lite2 == 1) glEnable(GL_LIGHT2);
	else glDisable(GL_LIGHT2);
	if(lite3 == 1) glEnable(GL_LIGHT3);
	else glDisable(GL_LIGHT3);
}

void drawships(void)
{
	int i;

	// Draws the ships, manipulating the properties as we go
	for(i=0; i < NUMSHIPS; i++)
	{
		// First, check if we wanna draw it...
		if(shipdb[i][7])
		{
			glPushMatrix();
				// Translate a little bit from saved values
				glTranslatef(shipdb[i][0], shipdb[i][1], shipdb[i][2]);
	
				// Rotate the axes from saved values
				glRotatef(shipdb[i][3], 1.0, 0.0, 0.0);
				glRotatef(-shipdb[i][4]+90, 0.0, 1.0, 0.0);
				glRotatef(shipdb[i][5], 0.0, 0.0, 1.0);
		
				// Small Star Destroyer
//				if(i == 0)
//					stardestroyer(shipdb[i][8]);
					
				// New Star Destroyer (#0)
				if(i == 0)
					drawextobj(2, shipdb[i][8]);
				// TIE Fighter
				else if(i == 1)
					tiefighter(shipdb[i][8]);
				// TIE Fighter Right Side
				else if(i == 2)
					tiefighter(shipdb[i][8]);
				// Imperial Shuttle
				else if(i == 3)
					shuttle(shipdb[i][8], shipdb[i][6]);
				// Y-Wing Fighter
				else if(i == 4)
					drawextobj(0, shipdb[i][8]);
				// TIE Bomber
				else if(i == 5)
					drawextobj(1, shipdb[i][8]);
				// New Star Destroyer
				else if(i == 6)
					drawextobj(2, shipdb[i][8]);
				// Millennium Falcon
				else if(i == 7)
					drawextobj(3, shipdb[i][8]);
			glPopMatrix();
		}
	}

}

// Main display function
void display(void)
{
	// Calculates and prints fps
	timestuff();

	// Set miscellaneous GL stuffs to get a nice camera view
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	if(selected_ship != -1 && bool_shipview == -1)
		gluPerspective(60,ratio,.01,100);
	else
		gluPerspective(60,ratio,.01,100);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// Update hte position and stuff
	updatevalues();

	// Updates the camera location smoothly
	camerasmooth();

	// Actually draws the ships
	drawships();

	// First draw a big sphere which will be used for a starfield
	setMaterials(&whiteboy, GL_FRONT_AND_BACK);
	fakestarfield();

	glutSwapBuffers();
}

void resize(GLsizei w, GLsizei h)
{
	// Set the global width and height
	win_width = w;
	win_height = h;

	// Sets the ratio of the resized window
	ratio = (GLfloat)w/(GLfloat)h;

	// Set the viewport to take advantage of the new window dimensions
	glViewport(0,0,(GLsizei)win_width,(GLsizei)win_height);
}

// generic initialization routine
void myinit(void)
{
	// Set up all the light properties and positions
	GLfloat light0_ambient[] = { 1.0, 1.0, 1.0, 1.0 };
	GLfloat light0_position[] = { 0.0, 2.0, 0.0, 0.0 };

	GLfloat light1_ambient[] = { 0.0, 0.0, 1.0, 1.0 };
	GLfloat light1_position[] = { 1.0, 0.0, 0.0, 0.0 };
	GLfloat light1_diffuse[] = { 0.0, 0.0, 1.0, 1.0 };

	GLfloat light2_ambient[] = { 1.0, 0.0, 0.0, 1.0 };
	GLfloat light2_position[] = { -1.0, 0.0, 0.0, 0.0 };
	GLfloat light2_diffuse[] = { 1.0, 0.0, 0.0, 1.0 };

	GLfloat light3_ambient[] = { 0.0, 1.0, 0.0, 1.0 };
	GLfloat light3_position[] = { 0.0, -1.0, 0.0, 0.0 };
	GLfloat light3_diffuse[] = { 0.0, 1.0, 0.0, 1.0 };

	// Some borrowed code (next two lines)
	LoadGLTextures();                           // Load The Texture(s) 

	glEnable(GL_AUTO_NORMAL);

	// Initial conditions for the ships/objects
	init_ships();

	// Make a new quadric object which we will use
	gluObj = gluNewQuadric();

	glClearColor(0.0, 0.0, 0.0, 0.0);
	glClearDepth(100.0);
	glShadeModel(GL_SMOOTH);
	
	// Apply the properties to the lights
	glLightfv(GL_LIGHT0, GL_AMBIENT, light0_ambient);
	glLightfv(GL_LIGHT0, GL_POSITION, light0_position);

	glLightfv(GL_LIGHT1, GL_AMBIENT, light1_ambient);
	glLightfv(GL_LIGHT1, GL_POSITION, light1_position);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_diffuse);
	glLightfv(GL_LIGHT2, GL_AMBIENT, light2_ambient);
	glLightfv(GL_LIGHT2, GL_POSITION, light2_position);
	glLightfv(GL_LIGHT2, GL_DIFFUSE, light2_diffuse);
	glLightfv(GL_LIGHT3, GL_AMBIENT, light3_ambient);
	glLightfv(GL_LIGHT3, GL_POSITION, light3_position);
	glLightfv(GL_LIGHT3, GL_DIFFUSE, light3_diffuse);

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);

	// Play with Alpha Blending
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);

//	glutSwapBuffers();
}

void root_menu(int id)
{
	// Dummy Function
	if(id == 1)
		exit(0);
}

// If we type a key on the keyboad
void key(unsigned char k, int x, int y)
{
	// Keybindings
	if(k == 'q') exit(0);

	// Cycles through filled, wire, and point drawstyles
	if(k == 'p')
	{
		if(drawstyle < 2) drawstyle++;
		else drawstyle=0;

		if(!drawstyle)
			glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		else if(drawstyle == 1)
			glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		else if(drawstyle == 2)
			glPolygonMode(GL_FRONT_AND_BACK, GL_POINT);
	}

	// Tabs through the ships (TAB == 't' temporarily...)
	if(k == 't')
	{
		if(selected_ship == (NUMSHIPS-1))
			selected_ship = -1;
		else
			selected_ship++;

		infoships(selected_ship);

		if((selected_ship == 0) || (selected_ship == 6))
			follow_zoom=2.5;
		else
			follow_zoom=0.75;
		
		if(debug)
		{
			if(selected_ship == -1)
				printf("Debug: selected_ship = (None)\n");
			else
				printf("Debug: selected_ship = %d\n", selected_ship);
		}
	}

	// Toggles ship view between follow and centerlook
	// If not in orbit view
	if((selected_ship != -1) && (k == 'm'))
		bool_shipview *= -1;

	if(k == 'a') bool_movey = 1;
	if(k == 'z') bool_movey = -1;

}

// Keybindings for ones that turn off when unpressed
void keyup(unsigned char k, int x, int y)
{
	// Keybindings up
	if(k == 'a') bool_movey = 0;
	if(k == 'z') bool_movey = 0;
}

// special keys
void skey(unsigned char k, int x, int y)
{
	if(selected_ship == -1)
	{
		// Manage the whole rotation
		if(k == GLUT_KEY_LEFT) orbit_xang_bool = -1;
		if(k == GLUT_KEY_RIGHT) orbit_xang_bool = 1;
		if(k == GLUT_KEY_UP) orbit_yang_bool = 1;
		if(k == GLUT_KEY_DOWN) orbit_yang_bool = -1;
		if(k == GLUT_KEY_HOME) orbit_zoom_bool = -1;
		if(k == GLUT_KEY_END) orbit_zoom_bool = 1;
	}
	else
	{
		// Manage the whole rotation
		if(k == GLUT_KEY_LEFT) bool_angy = -1;
		if(k == GLUT_KEY_RIGHT) bool_angy = 1;
		if(k == GLUT_KEY_UP) bool_movez = 1;
		if(k == GLUT_KEY_DOWN) bool_movez = -1;
		if(k == GLUT_KEY_HOME) bool_follow_zoom = -1;
		if(k == GLUT_KEY_END) bool_follow_zoom = 1;
	}

	// bunch of Fx keys toggle lights
	if(k == GLUT_KEY_F1) lite0 *= -1;
	if(k == GLUT_KEY_F2) lite1 *= -1;
	if(k == GLUT_KEY_F3) lite2 *= -1;
	if(k == GLUT_KEY_F4) lite3 *= -1;
}

// Up versions of the above bindings
void skeyup(unsigned char k, int x, int y)
{
	if(selected_ship == -1)
	{
		// Manage the whole rotation
		if(k == GLUT_KEY_LEFT) orbit_xang_bool = 0;
		if(k == GLUT_KEY_RIGHT) orbit_xang_bool = 0;
		if(k == GLUT_KEY_UP) orbit_yang_bool = 0;
		if(k == GLUT_KEY_DOWN) orbit_yang_bool = -0;
		if(k == GLUT_KEY_HOME) orbit_zoom_bool = 0;
		if(k == GLUT_KEY_END) orbit_zoom_bool = 0;
	}
	else
	{
		// Manage the whole rotation
		if(k == GLUT_KEY_LEFT) bool_angy = 0;
		if(k == GLUT_KEY_RIGHT) bool_angy = 0;
		if(k == GLUT_KEY_UP) bool_movez = 0;
		if(k == GLUT_KEY_DOWN) bool_movez = -0;
		if(k == GLUT_KEY_HOME) bool_follow_zoom = 0;
		if(k == GLUT_KEY_END) bool_follow_zoom = 0;
	}
}

// Main function
int main(int argc, char **argv)
{
	ratio = (GLfloat)win_width/(GLfloat)win_height;

	if((argc>1) && !strcmp(argv[1], "-debug"))
		debug=1;

	// GLUT Init Stuff
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutInitWindowSize(INIT_W, INIT_H);
	glutInitWindowPosition(128,128);
	glutIdleFunc(display);

	// Reads in external object (RAW) files
	readextobjs();
		
	// Main Window
	glutCreateWindow("Star Wars VII - The OpenGL Strikes Back");
	glutDisplayFunc(display);
	glutReshapeFunc(resize);

	// Keys
	glutIgnoreKeyRepeat(1);
	glutKeyboardFunc(key);
	glutKeyboardUpFunc(keyup);
	glutSpecialFunc((void *)skey);
	glutSpecialUpFunc((void *)skeyup);

	// Lame Menu
	glutCreateMenu(root_menu);
	glutCreateMenu(root_menu);
	glutAddMenuEntry("Exits", 1);
	glutAttachMenu(GLUT_RIGHT_BUTTON);

	myinit();

	glutMainLoop();

	return 0;
}
